from .methods import methods
from .nodes import node_classes, Media, User, Arg

class Task(dict):
    """"
    task:

        nodes: [node1, node2] # these are all Node instances

        edges:
            -   type:   feed

            -   type:   send
                args:
                    messages = []
                    stuff =    whokonws

        nodes: [url1, url2]

        edges:
            -   type: upload
                args:
                    caption: 'text bla bla'
                    location: id
                    usertags: id

    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


def make_task(body):

    nodes = []
    edges = []
    edges = []


    if 'nodes' in body:
        nodes += body['nodes']

    if 'edges' in body:
        for edge in body['edges']:

            if isinstance(edge, dict):
                type = list(edge.keys())[0]
                args = list(edge.values())[0]
                edges += [dict(type=type, args=args)]
            else:
                edges += [dict(type=edge, args={})]
    else:
        raise Exception('in every action there must be actions')

    edges += [dict(type='evaluate', args={})]

    if 'name' in body:
        edges[0]['name'] = body['name']

    nodes = initialize_nodes(nodes, edges, body)
    return Task(nodes=nodes, edges=edges)


# TODO this shit suppress my errors
def partitionate(task: Task, bots):

    couples = []

    for (partition, bot) in enumerate(bots):

        def _right_partition(i):
            return i % len(bots) == partition

        new_nodes = [node for (i, node) in enumerate(task.nodes) if _right_partition(i)]

        new_edges = [dict(**edge) for edge in task.edges]
        new_task = Task(nodes=new_nodes, edges=new_edges)


        couples += [(new_task, bot)]

    return couples


def popped(to_pop, dictionary):
    return {key:value for key, value in dictionary.items() if key != to_pop}



def initialize_nodes(nodes, edges, data):
    if 'from_type' in data:
        Class = node_classes[data['from_type']]
    else:
        first_method = methods.get(edges[0]['type'], None)
        if not first_method:
            raise Exception('can\'t find {} edge in available edges methods')
        Class = first_method.accepts
        Class = Class if Class.__name__ != 'Node' else \
            Media if 'instagram.com' in nodes[0] else Arg

    return [Class(generic=value) for value in nodes]



from typing import List
from funcy import  rcompose, raiser, ignore
import time
from .common import accepts, today, tap
from ..nodes import Node, User, Media




@accepts(Media)
def like(bot, nodes,  args):


    max = float(args['max']) if 'max' in args else 1

    count = 0

    def increment():
        bot.total['likes'] += 1
        nonlocal count
        count += 1

    stop = raiser(StopIteration)

    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        # lambda node: node \
        #     if bot.suitable(node) \
        #     else tap(None,lambda: bot.logger.warn('{} not suitable'.format(node))),
        lambda node: like_media(node, bot=bot) \
            if node else None,
        lambda x: tap(x, increment) if x else None,
    )


    liked = map(process, nodes)
    liked = filter(lambda x: x, liked)

    return liked, bot.last


def like_media(media, bot):
        bot.api.like(media.id)
        if bot.last['status'] != 'ok':
            bot.logger.warn('request didn\'t return "ok" liking {}'.format(media.url))
            bot.sleep('error')
            return None
        else:
            with bot.cache as cache:
                cache['liked'].insert(
                    dict(identifier=media.id,
                        specifier=media.url,
                        time=today(),
                        type='media',
                        )
                )
            bot.logger.debug('liked %s' % media.url)
            bot.sleep('like')
            return media



from .node import Node
from .common import attributes


"""
{
	id
	latest_reel_media
	expiring_at
	seen
	can_reply
	can_reshare
	reel_type
	user {
		pk
		username
		full_name
		is_private
		profile_pic_url
		profile_pic_id
		friendship_status {
			following
			followed_by
			blocking
			muting
			is_private
			incoming_request
			outgoing_request
			is_bestie
		}
		is_verified
	}
	items[9] {
		taken_at
		pk
		id
		device_timestamp
		media_type
		code
		client_cache_key
		filter_type
		image_versions2 {
			candidates[2] {
				width
				height
				url
			}
		}
		original_width
		original_height
		caption_position
		is_reel_media
		is_dash_eligible
		video_dash_manifest
		video_codec
		number_of_qualities
		video_versions[3] {
			type
			width
			height
			url
			id
		}
		has_audio
		video_duration
		user {
			pk
			username
			full_name
			is_private
			profile_pic_url
			profile_pic_id
			is_verified
			has_anonymous_profile_picture
			reel_auto_archive
			is_unpublished
			is_favorite
		}
		caption {
			_
		}
		caption_is_edited
		photo_of_you
		can_viewer_save
		organic_tracking_token
		expiring_at
		imported_taken_at
		can_reshare
		story_hashtags[1] {
			x
			y
			z
			width
			height
			rotation
			is_pinned
			is_hidden
			hashtag {
				name
				id
			}
		}
		supports_reel_reactions
		show_one_tap_fb_share_tooltip
		has_shared_to_fb
	}
	prefetch_count
	has_besties_media
	media_count
	status
}
"""



class Story(Node):

    def __init__(self, *, generic=None, id=None, data=None,):
        self._id = id
        self._data = data

        if generic:
            self._data = generic

    def __repr__(self):
            url = self.get_url()
            return 'Story(id=\'{}\', url=\'{}\')'.format(self._id, url)

    def get_data(self, bot):
        return self._data

    def get_expiring_at(self):
        data = self._data
        if 'expiring_at' in data:
            return data['expiring_at']
        else:
            return None

    def get_taken_at(self):
        data = self._data
        if 'taken_at' in data:
            return data['taken_at']
        else:
            return None

    def get_url(self):
        data = self._data
        if 'video_versions' and 'video_duration' in data:
            return data['video_versions'][0]['url']
        elif 'image_versions2' in data:
            return data['image_versions2']['candidates'][0]['url']
        else:
            return None



from typing import List
import requests
from random import random
from pathlib import Path
from .common import accepts, today
from ..nodes import Node, Arg, Media
import time




@accepts(Media)
def delete(bot, nodes,  args):


    for node in nodes:

        if 'instagram.com' in node.url:
            try:
                media_id = node.id
                bot.api.media_info(media_id)
                media = bot.last['items'][0]
                if not bot.api.delete_media(media):
                    bot.logger.warn('deletion didn\'t go well')
            except (KeyError, TypeError):
                bot.logger.warn('cannot delete media {}'.format(node.url))
                pass

        else:
            raise Exception('`delete` needs urls of photos to delete')

        bot.logger.debug('sleeping some time')
        bot.sleep('delete')


    return [], bot.last



from .node import Node
from .geotag import Geotag
from .hashtag import Hashtag
import time




class User(Node):

    def __init__(self, *, generic=None, id=None, username=None, is_private=None, data={}):
        self._username = username
        self._id = id
        self._data = data

        if generic:
            self._username = generic

    def __repr__(self):
        username, id, data = attributes(self)

        if username:
            return 'User(username=\'{}\')'.format(username)
        elif id:
            return 'User(id=\'{}\')'.format(id)
        elif data:
            return 'User(data=\'{...}\')'

    @property
    def username(self):
        username, id, data = attributes(self)
        if username:
            return username
        elif data:
            return data['username']
        else:
            return None

    @property
    def id(self):
        username, id, data, *rest = attributes(self)
        if id:
            return id
        elif data:
            return data['pk']
        else:
            return None


    def get_data(self, bot):

        if self._id:
            bot.sleep('usual')
            bot.api.get_username_info(self._id)
            if 'user' in bot.last:
                self._data = bot.last['user']
                return self._data

        elif self._username:
            bot.sleep('usual')
            bot.api.search_username(self._username)
            if 'user' in bot.last:
                bot.api.get_username_info(bot.last['user']['pk'])
                if 'user' in bot.last:
                    self._data = bot.last['user']
                    return self._data
            else:
                return {}
        elif self._data:
            data = self._data
            if 'pk' in data:
                self._id = data['pk']
                return self.get_data(bot)
            else:
                return {}
        else:
            return {}

    def get_username(self, bot):
        if self.username:
            return self.username
        else:
            data = self.get_data(bot)
            return data['username']


    def get_id(self, bot):
        username, id, data = attributes(self)
        if id:
            return id
        elif data:
            if 'pk' in data:
                return data['pk']
            else:
                data = self.get_data(bot)
                return data['pk'] if 'pk' in data else None
        elif username:
            data = self.get_data(bot)
            return data['pk'] if 'pk' in data else None
        else:
            return None


    def get_followers_count(self, bot):
        _, _, data = attributes(self)
        if 'follower_count' in data:
            return data['follower_count']
        else:
            data = self.get_data(bot)
            return data['follower_count']


    def get_following_count(self, bot):
        _, _, data = attributes(self)
        if 'following_count' in data:
            return data['following_count']
        else:
            data = self.get_data(bot)
            return data['following_count']


    def get_is_private(self, bot):
        _, _, data = attributes(self)
        if 'is_private' in data:
            return data['is_private']
        else:
            data = self.get_data(bot)
            return data['is_private']


    def get_is_business(self, bot):
        _, _, data = attributes(self)
        if 'is_business' in data:
            return data['is_business']
        else:
            data = self.get_data(bot)
            return data['is_business']

    def get_is_verified(self, bot):
        _, _, data = attributes(self)
        if 'is_verified' in data:
            return data['is_verified']
        else:
            data = self.get_data(bot)
            return 'is_verified' in data and data['is_verified']

    def get_is_anonymous_picture(self, bot):
        _, _, data = attributes(self)
        if 'has_anonymous_profile_picture' in data:
            return data['has_anonymous_profile_picture']
        else:
            data = self.get_data(bot)
            return data['has_anonymous_profile_picture']

    def get_bio(self, bot):
        _, _, data = attributes(self)
        if 'biography' in data:
            return data['biography']
        else:
            data = self.get_data(bot)
            return data['biography']



def attributes(node):
    return  node._username, node._id, node._data



from funcy import rcompose, mapcat
from typing import List
from ..nodes import User, Media
from .common import accepts


@accepts(Media)
def usertags(bot, nodes,  args) -> List[User]:

    process = lambda media: media.get_usertags(bot)
    result = mapcat(process, nodes)

    return result, bot.last



import hashlib
import hmac
import json
import logging
import os
import sys
import time
import uuid
from random import uniform

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

import requests
import requests.utils
import urllib
from tqdm import tqdm

from . import config, devices
from .api_photo import configure_photo, download_photo, upload_photo
from .api_video import configure_video, download_video, upload_video
from .prepare import delete_credentials, get_credentials

from colorlog import ColoredFormatter


PY2 = sys.version_info[0] == 2

class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, prefix):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        return '[%s] %s' % (self.prefix, msg), kwargs

class API(object):

    _id = 0

    def __init__(self, logs_file, device=None, username=None,  id=None):

        self._id = API._id if not id else id
        API._id += 1
        # Setup device and user_agent
        device = device or devices.DEFAULT_DEVICE
        self.device_settings = devices.DEVICES[device]
        self.user_agent = config.USER_AGENT_BASE.format(**self.device_settings)

        self.is_logged_in = False
        self.last_response = None
        self.total_requests = 0

        # Setup logging
        self.logger = logging.getLogger('[{}]'.format(self._id))

        # fh = HTMLFileHandler(title=self._id, file=logs_file, mode='w')
        # fh.setLevel(logging.INFO)
        # fh.setFormatter(file_formatter())
        # self.logger.addHandler(fh)

        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(colred_formatter())
        self.logger.addHandler(ch)

        self.logger.setLevel(logging.DEBUG)
        self.logger = LoggerAdapter(self.logger, username)



        self.last_json = None



    def media_info(self, media_id):
        url = 'media/{media_id}/info/'.format(media_id=media_id)
        return self.send_request(url)


    def send_request(self, endpoint, post=None, login=False, with_signature=True):
            self.logger.info('new request to endpoint %s' % endpoint)
            if (not self.is_logged_in and not login):
                msg = "Not logged in!"
                self.logger.critical(msg)
                raise Exception(msg)

            self.session.headers.update(config.REQUEST_HEADERS)
            self.session.headers.update({'User-Agent': self.user_agent})
            try:
                self.total_requests += 1
                if post is not None:  # POST
                    if with_signature:
                        # Only `send_direct_item` doesn't need a signature
                        post = self.generate_signature(post)
                    response = self.session.post(
                        config.API_URL + endpoint, data=post)
                else:  # GET
                    response = self.session.get(
                        config.API_URL + endpoint)
            except Exception as e:
                self.logger.warning(str(e))
                return False

            if response.status_code == 200:
                self.last_response = response
                try:
                    self.last_json = json.loads(response.text)
                    return True
                except JSONDecodeError:
                    return False
            else:
                self.logger.error("Request returns {} error!".format(response.status_code))
                if response.status_code == 429:
                    sleep_minutes = 5
                    self.logger.warning(
                        "That means 'too many requests'. I'll go to sleep "
                        "for {} minutes.".format(sleep_minutes))
                    time.sleep(sleep_minutes * 60)
                elif response.status_code == 400:
                    response_data = json.loads(response.text)
                    msg = "Instagram's error message: {}"
                    self.logger.info(msg.format(response_data.get('message')))
                    if 'error_type' in response_data:
                        msg = 'Error type: {}'.format(response_data['error_type'])
                        self.logger.info(msg)

                # For debugging
                try:
                    self.last_response = response
                    self.last_json = json.loads(response.text)
                except Exception:
                    pass
                return False

    def edit_profile(self, external_url, phone_number, full_name, biography, email, gender, **rest):
        data = self.json_data({
            'external_url': external_url,
            'phone_number': phone_number,
            'username': self.username,
            'full_name': full_name,
            'biography': biography,
            'email': email,
            'gender': gender,
        })
        return self.send_request('accounts/edit_profile/', data)

    def get_story_feed(self, user_id):
        """
        last_json will have the form:
        {
            reel:
                items: {
                ...
                }
                id
                user
                expiring_at
                location
                reel_type
                title
                ...
            }
        }
        """
        url = "feed/user/{user_id}/story/".format(user_id=user_id)
        return self.send_request(url)

    def get_user_reel(self, user_id):
        url = 'feed/user/{}/reel_media/'.format(user_id)
        return self.send_request(url)

    def get_user_stories(self, user_id):
        url = 'feed/user/{}/story/'.format(user_id)
        return self.send_request(url)

    def get_self_story_viewers(self, story_id):

        url = 'media/{}/list_reel_media_viewer/?supported_capabilities_new={}'.format(story_id,
                                                                                      config.SUPPORTED_CAPABILITIES)
        return self.send_request(url)

    def get_tv_suggestions(self):
        url = 'igtv/tv_guide/'
        return self.send_request(url)

    def get_hashtag_stories(self, hashtag):
        url = 'tags/{}/story/'.format(hashtag)
        return self.send_request(url)

    def follow_hashtag(self, hashtag):
        data = self.json_data({})
        url = 'tags/follow/{}/'.format(hashtag)
        return self.send_request(url, data)

    def unfollow_hashtag(self, hashtag):
        data = self.json_data({})
        url = 'tags/unfollow/{}/'.format(hashtag)
        return self.send_request(url, data)

    def get_tags_followed_by_user(self, user_id):
        url = 'users/{}/following_tags_info/'.format(user_id)
        return self.send_request(url)

    def get_hashtag_sections(self, hashtag):
        data = self.json_data({'supported_tabs': "['top','recent','places']", 'include_persistent': 'true'})
        url = 'tags/{}/sections/'.format(hashtag)
        return self.send_request(url, data)

    def get_media_insight(self, media_id):
        url = 'insights/media_organic_insights/{}/?ig_sig_key_version={}'.format(media_id, config.IG_SIG_KEY)
        return self.send_request(url)

    def get_self_insight(self):
        url = 'insights/account_organic_insights/?show_promotions_in_landing_page=true&first={}'.format()  # todo
        return self.send_request(url)

    def save_media(self, media_id):
        data = self.json_data()
        url = 'media/{}/save/'.format(media_id)
        return self.send_request(url, data)

    def unsave_media(self, media_id):
        data = self.json_data()
        url = 'media/{}/unsave/'.format(media_id)
        return self.send_request(url, data)

    def get_saved_medias(self):
        url = 'feed/saved/'
        return self.send_request(url)

    ########################################################################################################

    def set_user(self, username, password):
        self.username = username
        self.password = password
        self.uuid = self.generate_UUID(uuid_type=True)

    def login(self, username=None, password=None, force=False, proxy=None,
              use_cookie=False, cookie_fname='cookie.txt'):
        if password is None:
            username, password = get_credentials(username=username)

        self.device_id = self.generate_device_id(self.get_seed(username, password))
        self.proxy = proxy
        self.set_user(username, password)

        cookie_is_loaded = False
        if use_cookie:
            try:
                self.load_cookie(cookie_fname)
                cookie_is_loaded = True
                self.is_logged_in = True
                self.set_proxy()  # Only happens if `self.proxy`
                self.logger.info("Logged-in successfully as '{}' using the cookie!".format(self.username))
                return True
            except Exception:
                print("The cookie is not found, but don't worry `instabot`"
                      " will create it for you using your login details.")

        if not cookie_is_loaded and (not self.is_logged_in or force):
            self.session = requests.Session()
            self.set_proxy()  # Only happens if `self.proxy`
            url = 'si/fetch_headers/?challenge_type=signup&guid={uuid}'
            url = url.format(uuid=self.generate_UUID(False))
            if self.send_request(url, login=True):
                data = json.dumps({
                    'phone_id': self.generate_UUID(True),
                    '_csrftoken': self.token,
                    'username': self.username,
                    'guid': self.uuid,
                    'device_id': self.device_id,
                    'password': self.password,
                    'login_attempt_count': '0',
                })

                if self.send_request('accounts/login/', data, True):
                    self.is_logged_in = True
                    self.logger.info("Logged-in successfully as '{}'!".format(self.username))
                    if use_cookie:
                        self.save_cookie(cookie_fname)
                        self.logger.info("Saved cookie!")
                    return True
                else:
                    self.logger.info("Username or password is incorrect.")
                    delete_credentials()
                    return False

    def load_cookie(self, fname):
        # Python2 compatibility
        if PY2:
            FileNotFoundError = IOError

        try:
            with open(fname, 'r') as f:
                self.session = requests.Session()
                self.session.cookies = requests.utils.cookiejar_from_dict(json.load(f))
            cookie_username = self.cookie_dict['ds_user']
            assert cookie_username == self.username
        except FileNotFoundError:
            raise Exception('Cookie file `{}` not found'.format(fname))
        except (TypeError, EOFError):
            os.remove(fname)
            msg = ('An error occured opening the cookie `{}`, '
                   'it will be removed an recreated.')
            raise Exception(msg.format(fname))
        except AssertionError:
            msg = 'The loaded cookie was for {} instead of {}.'
            raise Exception(msg.format(cookie_username, self.username))

    def save_cookie(self, fname):
        with open(fname, 'w') as f:
            json.dump(requests.utils.dict_from_cookiejar(self.session.cookies), f)

    def logout(self, *args, **kwargs):
        if not self.is_logged_in:
            return True
        self.is_logged_in = not self.send_request('accounts/logout/')
        return not self.is_logged_in

    def set_proxy(self):
        if self.proxy:
            parsed = urllib.parse.urlparse(self.proxy)
            scheme = 'http://' if not parsed.scheme else ''
            self.session.proxies['http'] = scheme + self.proxy
            self.session.proxies['https'] = scheme + self.proxy


    @property
    def cookie_dict(self):
        return self.session.cookies.get_dict()

    @property
    def token(self):
        return self.cookie_dict['csrftoken']

    @property
    def user_id(self):
        return self.cookie_dict['ds_user_id']

    @property
    def rank_token(self):
        return "{}_{}".format(self.user_id, self.uuid)

    @property
    def default_data(self):
        return {
            '_uuid': self.uuid,
            '_uid': self.user_id,
            '_csrftoken': self.token,
        }

    def json_data(self, data=None):
        """Adds the default_data to data and dumps it to a json."""
        if data is None:
            data = {}
        data.update(self.default_data)
        return json.dumps(data)

    def sync_features(self):
        data = self.json_data({'id': self.user_id, 'experiments': config.EXPERIMENTS})
        return self.send_request('qe/sync/', data)

    def auto_complete_user_list(self):
        return self.send_request('friendships/autocomplete_user_list/')

    def get_timeline_feed(self):
        """ Returns 8 medias from timeline feed of logged user."""
        return self.send_request('feed/timeline/')

    def get_megaphone_log(self):
        return self.send_request('megaphone/log/')

    def expose(self):
        data = self.json_data({
            'id': self.user_id,
            'experiment': 'ig_android_profile_contextual_feed'
        })
        return self.send_request('qe/expose/', data)

    def upload_photo(self, photo, caption=None, upload_id=None, from_video=False):
        return upload_photo(self, photo, caption, upload_id, from_video)

    def download_photo(self, media_id, filename, media=False, folder='photos'):
        return download_photo(self, media_id, filename, media, folder)

    def configure_photo(self, upload_id, photo, caption=''):
        return configure_photo(self, upload_id, photo, caption)

    def upload_video(self, photo, caption=None, upload_id=None):
        return upload_video(self, photo, caption, upload_id)

    def download_video(self, media_id, filename, media=False, folder='video'):
        return download_video(self, media_id, filename, media, folder)

    def configure_video(self, upload_id, video, thumbnail, width, height, duration, caption=''):
        return configure_video(self, upload_id, video, thumbnail, width, height, duration, caption)

    def edit_media(self, media_id, captionText=''):
        data = self.json_data({'caption_text': captionText})
        url = 'media/{media_id}/edit_media/'.format(media_id=media_id)
        return self.send_request(url, data)

    def remove_self_tag(self, media_id):
        data = self.json_data()
        url = 'media/{media_id}/remove/'.format(media_id=media_id)
        return self.send_request(url, data)

    def media_info(self, media_id):
        # data = self.json_data({'media_id': media_id})
        url = 'media/{media_id}/info/'.format(media_id=media_id)
        return self.send_request(url)

    def archive_media(self, media, undo=False):
        action = 'only_me' if not undo else 'undo_only_me'
        data = self.json_data({'media_id': media['id']})
        url = 'media/{media_id}/{action}/?media_type={media_type}'.format(
            media_id=media['id'],
            action=action,
            media_type=media['media_type']
        )
        return self.send_request(url, data)

    def delete_media(self, media):
        data = self.json_data({'media_id': media.get('id')})
        url = 'media/{media_id}/delete/'.format(media_id=media.get('id'))
        return self.send_request(url, data)

    def change_password(self, new_password):
        data = self.json_data({
            'old_password': self.password,
            'new_password1': new_password,
            'new_password2': new_password
        })
        return self.send_request('accounts/change_password/', data)

    def explore(self):
        return self.send_request('discover/explore/')

    def comment(self, media_id, comment_text):
        data = self.json_data({'comment_text': comment_text})
        url = 'media/{media_id}/comment/'.format(media_id=media_id)
        return self.send_request(url, data)

    def reply_to_comment(self, media_id, comment_text, parent_comment_id):
        data = self.json_data({'comment_text': comment_text, 'replied_to_comment_id': parent_comment_id})
        url = 'media/{media_id}/comment/'.format(media_id=media_id)
        return self.send_request(url, data)

    def delete_comment(self, media_id, comment_id):
        data = self.json_data()
        url = 'media/{media_id}/comment/{comment_id}/delete/'
        url = url.format(media_id=media_id, comment_id=comment_id)
        return self.send_request(url, data)

    def get_username_info(self, user_id):
        url = 'users/{user_id}/info/'.format(user_id=user_id)
        return self.send_request(url)

    def get_self_username_info(self):
        return self.get_username_info(self.user_id)

    def get_recent_activity(self):
        return self.send_request('news/inbox/?')

    def get_following_recent_activity(self):
        return self.send_request('news/?')

    def getv2Inbox(self):
        return self.send_request('direct_v2/inbox/?')

    def get_user_tags(self, user_id):
        url = 'usertags/{user_id}/feed/?rank_token={rank_token}&ranked_content=true&'
        url = url.format(user_id=user_id, rank_token=self.rank_token)
        return self.send_request(url)

    def get_self_user_tags(self):
        return self.get_user_tags(self.user_id)

    def tag_feed(self, tag):
        url = 'feed/tag/{tag}/?rank_token={rank_token}&ranked_content=true&'
        return self.send_request(url.format(tag=tag, rank_token=self.rank_token))

    def get_comment_likers(self, comment_id):
        url = 'media/{comment_id}/comment_likers/?'.format(comment_id=comment_id)
        return self.send_request(url)

    def get_media_likers(self, media_id):
        url = 'media/{media_id}/likers/?'.format(media_id=media_id)
        return self.send_request(url)

    def get_geo_media(self, user_id):
        url = 'maps/user/{user_id}/'.format(user_id=user_id)
        return self.send_request(url)

    def get_self_geo_media(self):
        return self.get_geo_media(self.user_id)

    def sync_from_adress_book(self, contacts):
        url = 'address_book/link/?include=extra_display_name,thumbnails'
        return self.send_request(url, 'contacts=' + json.dumps(contacts))

    def get_timeline(self):
        url = 'feed/timeline/?rank_token={rank_token}&ranked_content=true&'
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_archive_feed(self):
        url = 'feed/only_me_feed/?rank_token={rank_token}&ranked_content=true&'
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_user_feed(self, user_id, max_id='', min_timestamp=None):
        url = 'feed/user/{user_id}/?max_id={max_id}&min_timestamp={min_timestamp}&rank_token={rank_token}&ranked_content=true'
        url = url.format(
            user_id=user_id,
            max_id=max_id,
            min_timestamp=min_timestamp,
            rank_token=self.rank_token
        )
        return self.send_request(url)

    def get_self_user_feed(self, max_id='', min_timestamp=None):
        return self.get_user_feed(self.user_id, max_id, min_timestamp)

    def get_hashtag_feed(self, hashtag, max_id=''):
        url = 'feed/tag/{hashtag}/?max_id={max_id}&rank_token={rank_token}&ranked_content=true&'
        url = url.format(
            hashtag=hashtag,
            max_id=max_id,
            rank_token=self.rank_token
        )
        return self.send_request(url)

    def get_location_feed(self, location_id, max_id=''):
        url = 'feed/location/{location_id}/?max_id={max_id}&rank_token={rank_token}&ranked_content=true&'
        url = url.format(
            location_id=location_id,
            max_id=max_id,
            rank_token=self.rank_token
        )
        return self.send_request(url)

    def get_popular_feed(self):
        url = 'feed/popular/?people_teaser_supported=1&rank_token={rank_token}&ranked_content=true&'
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_user_followings(self, user_id, max_id=''):
        url = 'friendships/{user_id}/following/?max_id={max_id}&ig_sig_key_version={sig_key}&rank_token={rank_token}'
        url = url.format(
            user_id=user_id,
            max_id=max_id,
            sig_key=config.SIG_KEY_VERSION,
            rank_token=self.rank_token
        )
        return self.send_request(url)

    def get_self_users_following(self):
        return self.get_user_followings(self.user_id)

    def get_user_followers(self, user_id, max_id=''):
        url = 'friendships/{user_id}/followers/?rank_token={rank_token}'
        url = url.format(user_id=user_id, rank_token=self.rank_token)
        if max_id:
            url += '&max_id={max_id}'.format(max_id=max_id)
        return self.send_request(url)

    def get_self_user_followers(self):
        return self.followers

    def like_comment(self, comment_id):
        data = self.json_data()
        url = 'media/{comment_id}/comment_like/'.format(comment_id=comment_id)
        return self.send_request(url, data)

    def unlike_comment(self, comment_id):
        data = self.json_data()
        url = 'media/{comment_id}/comment_unlike/'.format(comment_id=comment_id)
        return self.send_request(url, data)

    def like(self, media_id):
        data = self.json_data({'media_id': media_id})
        url = 'media/{media_id}/like/'.format(media_id=media_id)
        return self.send_request(url, data)

    def unlike(self, media_id):
        data = self.json_data({'media_id': media_id})
        url = 'media/{media_id}/unlike/'.format(media_id=media_id)
        return self.send_request(url, data)

    def get_media_comments(self, media_id, max_id=''):
        url = 'media/{media_id}/comments/'.format(media_id=media_id)
        if max_id:
            url += '?max_id={max_id}'.format(max_id=max_id)
        return self.send_request(url)

    def get_direct_share(self):
        return self.send_request('direct_share/inbox/?')

    def follow(self, user_id):
        data = self.json_data({'user_id': user_id})
        url = 'friendships/create/{user_id}/'.format(user_id=user_id)
        return self.send_request(url, data)

    def unfollow(self, user_id):
        data = self.json_data({'user_id': user_id})
        url = 'friendships/destroy/{user_id}/'.format(user_id=user_id)
        return self.send_request(url, data)

    def block(self, user_id):
        data = self.json_data({'user_id': user_id})
        url = 'friendships/block/{user_id}/'.format(user_id=user_id)
        return self.send_request(url, data)

    def unblock(self, user_id):
        data = self.json_data({'user_id': user_id})
        url = 'friendships/unblock/{user_id}/'.format(user_id=user_id)
        return self.send_request(url, data)

    def user_friendship(self, user_id):
        data = self.json_data({'user_id': user_id})
        url = 'friendships/show/{user_id}/'.format(user_id=user_id)
        return self.send_request(url, data)

    @staticmethod
    def _prepare_recipients(users, thread_id=None, use_quotes=False):
        if not isinstance(users, list):
            print('Users must be an list')
            return False
        result = {'users': '[[{}]]'.format(','.join(users))}
        if thread_id:
            template = '["{}"]' if use_quotes else '[{}]'
            result['thread'] = template.format(thread_id)
        return result

    def send_direct_item(self, item_type, users, **options):
        data = {
            'client_context': self.generate_UUID(True),
            'action': 'send_item'
        }

        url = 'direct_v2/threads/broadcast/{}/'.format(item_type)
        text = options.get('text', '')
        if item_type == 'link':
            data['link_text'] = text
            data['link_urls'] = json.dumps(options.get('urls'))
        elif item_type == 'text':
            data['text'] = text
        elif item_type == 'media_share':
            data['text'] = text
            data['media_type'] = options.get('media_type', 'photo')
            data['media_id'] = options.get('media_id', '')
        elif item_type == 'hashtag':
            data['text'] = text
            data['hashtag'] = options.get('hashtag', '')
        elif item_type == 'profile':
            data['text'] = text
            data['profile_user_id'] = options.get('profile_user_id')

        recipients = self._prepare_recipients(users, options.get('thread'), use_quotes=False)
        if not recipients:
            return False
        data['recipient_users'] = recipients.get('users')
        if recipients.get('thread'):
            data['thread_ids'] = recipients.get('thread')
        data.update(self.default_data)
        return self.send_request(url, data, with_signature=False)

    @staticmethod
    def generate_signature(data):
        body = hmac.new(config.IG_SIG_KEY.encode('utf-8'), data.encode('utf-8'),
                        hashlib.sha256).hexdigest() + '.' + urllib.parse.quote(data)
        signature = 'ig_sig_key_version={sig_key}&signed_body={body}'
        return signature.format(sig_key=config.SIG_KEY_VERSION, body=body)

    @staticmethod
    def generate_device_id(seed):
        volatile_seed = "12345"
        m = hashlib.md5()
        m.update(seed.encode('utf-8') + volatile_seed.encode('utf-8'))
        return 'android-' + m.hexdigest()[:16]

    @staticmethod
    def get_seed(*args):
        m = hashlib.md5()
        m.update(b''.join([arg.encode('utf-8') for arg in args]))
        return m.hexdigest()

    @staticmethod
    def generate_UUID(uuid_type):
        generated_uuid = str(uuid.uuid4())
        if uuid_type:
            return generated_uuid
        else:
            return generated_uuid.replace('-', '')

    def get_liked_media(self, max_id=''):
        url = 'feed/liked/?max_id={max_id}'.format(max_id=max_id)
        return self.send_request(url)

    def get_total_followers_or_followings(self,
                                          user_id,
                                          amount=None,
                                          which='followers',
                                          filter_private=False,
                                          filter_business=False,
                                          filter_verified=False,
                                          usernames=False,
                                          to_file=None,
                                          overwrite=False):
        from io import StringIO

        if which == 'followers':
            key = 'follower_count'
            get = self.get_user_followers
        elif which == 'followings':
            key = 'following_count'
            get = self.get_user_followings

        sleep_track = 0
        result = []
        next_max_id = ''
        self.get_username_info(user_id)
        username_info = self.last_json
        if "user" in username_info:
            total = amount or username_info["user"][key]

            if total > 200000:
                print("Consider temporarily saving the result of this big "
                      "operation. This will take a while.\n")
        else:
            return False
        if filter_business:
            print("--> You are going to filter business accounts. This will take time! <--")
            from random import random
        if to_file is not None:
            if os.path.isfile(to_file):
                if not overwrite:
                    print("File `{}` already exists. Not overwriting.".format(to_file))
                    return False
                else:
                    print("Overwriting file `{}`".format(to_file))
            with open(to_file, 'w'):
                pass
        desc = "Getting {} of {}".format(which, user_id)
        with tqdm(total=total, desc=desc, leave=True) as pbar:
            while True:
                get(user_id, next_max_id)
                last_json = self.last_json
                try:
                    with open(to_file, 'a') if to_file is not None else StringIO() as f:
                        for item in last_json["users"]:
                            if filter_private and item['is_private']:
                                continue
                            if filter_business:
                                time.sleep(2 * random())
                                self.get_username_info(item['pk'])
                                item_info = self.last_json
                                if item_info['user']['is_business']:
                                    continue
                            if filter_verified and item['is_verified']:
                                continue
                            if to_file is not None:
                                if usernames:
                                    f.write("{}\n".format(item['username']))
                                else:
                                    f.write("{}\n".format(item['pk']))
                            result.append(item)
                            pbar.update(1)
                            sleep_track += 1
                            if sleep_track >= 20000:
                                sleep_time = uniform(120, 180)
                                msg = "\nWaiting {:.2f} min. due to too many requests."
                                print(msg.format(sleep_time / 60))
                                time.sleep(sleep_time)
                                sleep_track = 0
                    if not last_json["users"] or len(result) >= total:
                        return result[:total]
                except Exception as e:
                    print("ERROR: {}".format(e))
                    return result[:total]

                if last_json["big_list"] is False:
                    return result[:total]

                next_max_id = last_json.get("next_max_id", "")

    def get_total_followers(self, user_id, amount=None):
        return self.get_total_followers_or_followings(
            user_id, amount, 'followers')

    def get_total_followings(self, user_id, amount=None):
        return self.get_total_followers_or_followings(
            user_id, amount, 'followings')

    def get_total_user_feed(self, user_id, min_timestamp=None):
        return self.get_last_user_feed(user_id, amount=float('inf'), min_timestamp=min_timestamp)

    def get_last_user_feed(self, user_id, amount, min_timestamp=None):
        user_feed = []
        next_max_id = ''
        while True:
            if len(user_feed) >= float(amount):
                # one request returns max 13 items
                return user_feed[:amount]
            self.get_user_feed(user_id, next_max_id, min_timestamp)
            last_json = self.last_json
            if 'items' not in last_json:
                return user_feed
            user_feed += last_json["items"]
            if not last_json.get("more_available"):
                return user_feed
            next_max_id = last_json.get("next_max_id", "")

    def get_total_hashtag_feed(self, hashtag_str, amount=100):
        hashtag_feed = []
        next_max_id = ''

        with tqdm(total=amount, desc="Getting hashtag media.", leave=False) as pbar:
            while True:
                self.get_hashtag_feed(hashtag_str, next_max_id)
                last_json = self.last_json
                if 'items' not in last_json:
                    return hashtag_feed[:amount]
                items = last_json['items']
                try:
                    pbar.update(len(items))
                    hashtag_feed += items
                    if not items or len(hashtag_feed) >= amount:
                        return hashtag_feed[:amount]
                except Exception:
                    return hashtag_feed[:amount]
                next_max_id = last_json.get("next_max_id", "")

    def get_total_self_user_feed(self, min_timestamp=None):
        return self.get_total_user_feed(self.user_id, min_timestamp)

    def get_total_self_followers(self):
        return self.get_total_followers(self.user_id)

    def get_total_self_followings(self):
        return self.get_total_followings(self.user_id)

    def get_total_liked_media(self, scan_rate=1):
        next_id = ''
        liked_items = []
        for _ in range(scan_rate):
            self.get_liked_media(next_id)
            last_json = self.last_json
            next_id = last_json.get("next_max_id", "")
            liked_items += last_json["items"]
        return liked_items

    def remove_profile_picture(self):
        data = self.json_data()
        return self.send_request('accounts/remove_profile_picture/', data)

    def set_private_account(self):
        data = self.json_data()
        return self.send_request('accounts/set_private/', data)

    def set_public_account(self):
        data = self.json_data()
        return self.send_request('accounts/set_public/', data)

    def set_name_and_phone(self, name='', phone=''):
        data = self.json_data({'first_name': name, 'phone_number': phone})
        return self.send_request('accounts/set_phone_and_name/', data)

    def get_profile_data(self):
        data = self.json_data()
        return self.send_request('accounts/current_user/?edit=true', data)

    def edit_profile(self, url, phone, first_name, biography, email, gender):
        data = self.json_data({
            'external_url': url,
            'phone_number': phone,
            'username': self.username,
            'full_name': first_name,
            'biography': biography,
            'email': email,
            'gender': gender,
        })
        return self.send_request('accounts/edit_profile/', data)

    def fb_user_search(self, query):
        url = 'fbsearch/topsearch/?context=blended&query={query}&rank_token={rank_token}'
        url = url.format(query=query, rank_token=self.rank_token)
        return self.send_request(url)

    def search_users(self, query):
        url = 'users/search/?ig_sig_key_version={sig_key}&is_typeahead=true&query={query}&rank_token={rank_token}'
        url = url.format(
            sig_key=config.SIG_KEY_VERSION,
            query=query,
            rank_token=self.rank_token
        )
        return self.send_request(url)

    def search_username(self, username):
        url = 'users/{username}/usernameinfo/'.format(username=username)
        return self.send_request(url)

    def search_tags(self, query):
        url = 'tags/search/?is_typeahead=true&q={query}&rank_token={rank_token}'
        url = url.format(query=query, rank_token=self.rank_token)
        return self.send_request(url)

    def search_location(self, query='', lat=None, lng=None):
        url = 'fbsearch/places/?rank_token={rank_token}&query={query}&lat={lat}&lng={lng}'
        url = url.format(rank_token=self.rank_token, query=query, lat=lat, lng=lng)
        return self.send_request(url)

    def get_user_reel(self, user_id):
        url = 'feed/user/{}/reel_media/'.format(user_id)
        return self.send_request(url)

    def get_user_stories(self, user_id):
        url = 'feed/user/{}/story/'.format(user_id)
        return self.send_request(url)

    def get_self_story_viewers(self, story_id):

        url = 'media/{}/list_reel_media_viewer/?supported_capabilities_new={}'.format(story_id,
                                                                                      config.SUPPORTED_CAPABILITIES)
        return self.send_request(url)

    def get_tv_suggestions(self):
        url = 'igtv/tv_guide/'
        return self.send_request(url)

    def get_hashtag_stories(self, hashtag):
        url = 'tags/{}/story/'.format(hashtag)
        return self.send_request(url)

    def follow_hashtag(self, hashtag):
        data = self.json_data({})
        url = 'tags/follow/{}/'.format(hashtag)
        return self.send_request(url, data)

    def unfollow_hashtag(self, hashtag):
        data = self.json_data({})
        url = 'tags/unfollow/{}/'.format(hashtag)
        return self.send_request(url, data)

    def get_tags_followed_by_user(self, user_id):
        url = 'users/{}/following_tags_info/'.format(user_id)
        return self.send_request(url)

    def get_hashtag_sections(self, hashtag):
        data = self.json_data({'supported_tabs': "['top','recent','places']", 'include_persistent': 'true'})
        url = 'tags/{}/sections/'.format(hashtag)
        return self.send_request(url, data)

    def get_media_insight(self, media_id):
        url = 'insights/media_organic_insights/{}/?ig_sig_key_version={}'.format(media_id, config.IG_SIG_KEY)
        return self.send_request(url)

    def get_self_insight(self):
        url = 'insights/account_organic_insights/?show_promotions_in_landing_page=true&first={}'.format()  # todo
        return self.send_request(url)

    def save_media(self, media_id):
        data = self.json_data()
        url = 'media/{}/save/'.format(media_id)
        return self.send_request(url, data)

    def unsave_media(self, media_id):
        data = self.json_data()
        url = 'media/{}/unsave/'.format(media_id)
        return self.send_request(url, data)

    def get_saved_medias(self):
        url = 'feed/saved/'
        return self.send_request(url)



def colred_formatter():
    format = '%(asctime)s | %(levelname)-8s | %(message)s'
    cformat = '%(log_color)s' + format
    date_format = '%Y-%m-%d %H:%M'
    return ColoredFormatter(cformat, date_format,
                            log_colors={'DEBUG': 'reset', 'INFO': 'green',
                                        'WARNING': 'yellow', 'ERROR': 'red',
                                        'CRITICAL': 'red'})

# def file_formatter():
#     format = '%(asctime)s | %(levelname)-8s | %(message)s'
#     cformat = '%(log_color)s' + format
#     date_format = '%Y-%m-%d %H:%M'
#     return HTMLFormatter(cformat, date_format,)



from .nodes import Media, User, Geotag, Hashtag

def make_predicate(script, bot):
    """
    filter:

        user:
            followers:      x > 50 and x < 1000
            following:      x < 500
            bio:            any(s not in ['porn', 'sex'] for s in x)
            is_private:     x and False
            is_business:    x and False
            is_verified:    x and False

        media:
            likes:          x < 200
            comments:       x < 100
            caption:        any(s not in x for s in ['porn', 'sex'])

        hashtag:
            name:           not in [porn, sex]

        geotag:
            name:           not in {{ cities }}
            lat:            x < 1000 and x > 500
            lng:            x < 1000 and x > 500


    """
    def check(lazy_expr, lazy_var):
        try:
            expr = lazy_expr()
            var = lazy_var()
            result = eval(expr, dict(x=var, bot=bot))
            # print("{} == {} for x={}".format(expr, result, var))
            return result

        except (KeyError, AttributeError):
            # bot.logger.error('failed to check expression')
            return True

    def predicate(node):


        bool = True

        if isinstance(node, Media):
            bool = bool and check(
                lambda: script['media']['likes'],
                lambda: node.get_like_count(bot)
            )
            bool = bool and check(
                lambda: script['media']['comments'],
                lambda: node.get_comment_count(bot)
            )
            bool = bool and check(
                lambda: script['media']['hashtags'],
                lambda: node.get_hashtags(bot)
            )
            bool = bool and check(
                lambda: script['media']['caption'],
                lambda: node.get_caption(bot)
            )
        elif isinstance(node, User):
            bool = bool and check(
                lambda: script['user']['id'],
                lambda: node.get_id(bot)
            )
            bool = bool and check(
                lambda: script['user']['followers'],
                lambda: node.get_followers_count(bot)
            )
            bool = bool and check(
                lambda: script['user']['following'],
                lambda: node.get_following_count(bot)
            )
            bool = bool and check(
                lambda: script['user']['bio'],
                lambda: node.get_bio(bot)
            )
            bool = bool and check(
                lambda: script['user']['is_private'],
                lambda: node.get_is_private(bot)
            )
            bool = bool and check(
                lambda: script['user']['is_business'],
                lambda: node.get_is_business(bot)
            )
            bool = bool and check(
                lambda: script['user']['is_verified'],
                lambda: node.get_is_verified(bot)
            )
        elif isinstance(node, Geotag):
            bool = bool and check(
                lambda: script['geotag']['name'],
                lambda: node.name
            )
            bool = bool and check(
                lambda: script['geotag']['lat'],
                lambda: node.get_coordinates(bot)[0])
            bool = bool and check(
                lambda: script['geotag']['lng'],
                lambda: node.get_coordinates(bot)[1]
            )
        elif isinstance(node, Hashtag):
            bool = bool and check(
                lambda: script['hashtag']['name'],
                lambda: node.name
            )

        else:
            return True

        if not bool:
            bot.logger.debug('{} not suitable, very sorry'.format(node))

        return bool

    return predicate



from funcy import  ignore
from functools import reduce
import time
import traceback

from .bot import Bot
from .methods import methods
from .threads import Thread

class Dont_retry(Exception):
    """
    this exception doesn't cause a retry of the edge method when it is raised.
    Useful when a method is not found in the methods object,
    because it would be just a waste of time to retry.
    """
    pass

class Reducer(Thread):

    def __init__(self, state, edges, name=None):
        super().__init__(name=state['bot'].username)
        self.name = state['bot'].username
        self.logger = state['bot'].logger
        self.edges = edges
        self.state = state



    def run(self):
        get_name = lambda: self.edges[0]['name']
        name = ignore((KeyError, AttributeError), 'unnamed')(get_name)()
        self.logger.info('beginning {} action'.format(name))
        last_state = reduce(_reducer, self.edges, self.state)
        super().set_data(last_state['data'])
        return




def _reducer(state: dict, edge: dict):

    nodes = state['nodes']
    bot: Bot = state['bot']
    errors = state['errors']
    data = state['data']

    type = edge['type']
    args = edge['args']

    if len(errors) > 1:
        # tried multiple times
        bot.logger.error('trying to solve errors: {}, bot: {}'.format(errors, bot))
        # send_bot_to_phone_verifier
        # change_bot_if_neccessary
        # resolve_captcha_if_necessary
        # sleep_more
        # remove_bot_if_broken

    try:

        # if not nodes:
        #     raise Dont_retry('no nodes, {}'.format(nodes))

        method = methods.get(type, None)

        if not method:
            raise Dont_retry('can\'t find method {}'.format(type))

        # bot.logger.debug('reducing nodes %s' % list(nodes))

        next_nodes, next_data = method(bot, nodes,  args)

        # bot.logger.info('{} did success on {}'.format(type, nodes))
        # bot.logger.debug('{} returned {}'.format(type, next_nodes))

        next_data = merge(data, {'__{}__'.format(type): next_data})

        # secs = bot.delay[type] if type in bot.delay else bot.delay['usual']
        # time.sleep(secs)

    except Dont_retry as exc:
        bot.logger.error('error reducing edge {}: \"{}\" {}'.format(type, exc.__class__.__name__, exc))
        return dict(nodes=[], bot=bot, errors=errors + [exc], data=data)

    except Exception as exc:
        bot.logger.error('error reducing edge {}: \"{}\" \n {}'.format(
            type,
            exc.__class__.__name__,
            '\n'.join(traceback.format_exc().split('\n')[5:])))
        bot.sleep('error')

        errored_state = merge(state, dict(errors=errors + [exc]))
        return _reducer(errored_state, edge)

    else:
        # all is right, no exceptions
        return dict(nodes=next_nodes, bot=bot, errors=[], data=next_data)


def merge(a, b):
    return { key: val for (key, val) in [*a.items(), *b.items()]}



from funcy import partial
from .nodes import Media, User, Geotag, Hashtag
from .make_predicate import make_predicate
from .bot import Bot





def make_bots(script):
    """


    bots:
        -
            username:           '{username}'
            password:           qwerty
            cache:              ./cache.db
            log:                ./logs.html
            cookie:             ./cookie.json
        -
            username:           giovanotti
            password:           qwerty
            cache:              ./cache.db
            log:                ./logs.html
            cookie:             ./cookie.json

    max_per_day:
        likes:                  50
        follow:                 20
        unfollow:               10
        ...

    delay:
        like:                   10
        usual:                  2
        ...

    """

    bots = []

    for data in script['bots']:
        params = dict(
            cache_file=data['cache'] if 'cache' in data else None,
            logs_file=data['logs'] if 'logs' in data else \
                data['log'] if 'log' in data else None,
            cookie_file=data['cookie'] if 'cookie' in data else None,
            username=data['username'] if 'username' in data \
                else error(Exception('username necessary')),
            password=data['password'] if 'password' in data \
                else error(Exception('password necessary')),
        )
        bots += [Bot(**params)]

    for bot in bots:

        if 'max_per_day' in script:
            bot.max_per_day = {key: value for key,
                               value in script['max_per_day'].items()}
        if 'delay' in script:
            bot.delay = {**bot.delay, **{key: value for key, value in script['delay'].items()} }

        if 'filter' in script:
            bot.predicates += [make_predicate(script['filter'], bot)]

    return bots

def error(exception):
    raise exception



from .common import accepts
from ..nodes import Node, User, Media, Story, Geotag, Hashtag

from .common import today, tap, dotdict
from ..bot import Bot
from dataset import connect
from funcy import rcompose, ignore, retry, mapcat
import time



@accepts(Node)
def scrape(bot: Bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else 1
        database = args['database']
        table = args['table']
        model = args['model']

    except:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}


    count = 0

    def increment():
        nonlocal count
        count += 1

    lazy_database = lambda: connect(
        database,
        engine_kwargs = {'connect_args': {'check_same_thread' : False}}
    )

    def process(node):
        """
        model:
            name:      x.full_name
            id:        x.pk
            followers: x.followers_count
        """
        with lazy_database() as db:
            insertion = dotdict()
            for name, expr in model.items():
                insertion[name] = evaluate(expr, node, bot=bot)

            if count <= max:
                db[table].insert(insertion)
                bot.logger.info('added to database node {} with insertion {}'.format(node, insertion))
                increment()
                yield node

            else:
                 return


    nodes = mapcat(process, nodes)


    return nodes, bot.last



def evaluate(expr, node, bot):
    x = node._data or node.get_data(bot)
    x = dotdict(**x)
    value = xeval(expr, x)
    if not value:
        x = node.get_data(bot)
        x = dotdict(**x)
        return xeval(expr, x)
    else:
        return value

def xeval(expr, x):
    try:
        return eval(expr, dict(x=x,
            User=User,
            Story=Story,
            Media=Media,
            Hashtag=Hashtag,
            Geotag=Geotag
        ))

    except (KeyError, AttributeError):
        return None



from collections import OrderedDict
from types import FunctionType


# returns the attributes of the class,
# TODO, in pyton 3.2 attribtes aren't ordered
def attributes(instance):
    list = [
        v
        for k, v in instance.__dict__.items()
        if not k.startswith('__') and not k.endswith('__')
        and not isinstance(v, (FunctionType, classmethod, staticmethod))
        ]
    return tuple(list)



from typing import List
import requests
from random import random
from pathlib import Path
from funcy import take
from itertools import islice
from .common import accepts, today
from ..nodes import Node, Arg, Media
import time




@accepts(Arg)
def upload(bot, nodes,  args):

    max = args.get('max') or 1
    nodes = islice(nodes, max)
    nodes = list(nodes)
    if len(nodes) == 0:
        bot.logger.error('no photos to upload')




    if max == 1:
        type = 'photo'

        node = take(1, nodes)[0]

        if isinstance(node, Arg):
            value = node.value
        elif isinstance(node, Media):
            value = node.url
        else:
            value = node

        caption = args['caption'] if 'caption' in args else None
        # TODO implement geotag and usertag in uploading
        # geotag = args[geotag] if 'geotag' in args else None
        # usertags = args[usertags] if 'usertags' in args else None



        if 'instagram.com' in value:
            media_id = Media(url=value).id
            bot.api.media_info(media_id)
            media = bot.last['items'][0]
            type = media_type(media)
            if type == 'photo':
                temp = TempImage(Path(bot.cache_file).parent)
                path = bot.api.download_photo(media_id, temp.path, media=media)
            else:
                temp = TempVideo(Path(bot.cache_file).parent)
                path = bot.api.download_video(media_id, temp.path, media=media)

        elif value.startswith(('http', 'https')):
            url = value
            if url.endswith(('.jpg', '.jpeg')):
                type = 'photo'
                temp = TempImage(Path(bot.cache_file).parent)
            elif url.endswith(('.mp4', '.mov')):
                type = 'video'
                temp = TempVideo(Path(bot.cache_file).parent)
            else:
                raise Exception('to use http url it must end with .jpg or .mp4 or .mov to notify file type')
            path = download_media(bot, url, temp.path)

        else:
            path = str(Path(value).resolve())
            if path.endswith(('.jpg', '.jpeg')):
                type = 'photo'
            elif path.endswith(('.mp4', '.mov')):
                type = 'video'
            else:
                raise Exception('media must be .jpg or .mp4 or .mov')




        if type == 'photo':
            if not bot.api.upload_photo(path, caption):
                bot.logger.warn('upload failed')
            else:
                bot.total['uploads'] += 1
        else:
            if not bot.api.upload_video(path, caption):
                bot.logger.warn('upload failed')

        temp.delete()

        # with bot.cache as cache:
        #     cache['uploaded'].insert(dict(identifier=bot.last, url=media.url, time=today(), type='media'))

        bot.logger.debug('sleeping some time')
        bot.sleep('upload')

    else:
        # upload album
        pass

    return [], bot.last


class TempImage():
    def __init__(self, folder):
        self.filename = str(random())[2:] + '.jpg'
        self._path = Path(folder).resolve() / self.filename
        self._path.exists() or self._path.touch()

    def delete(self):
        self._path.unlink()

    @property
    def path(self):
        return str(self._path)


class TempVideo():
    def __init__(self, folder):
        self.filename = str(random())[2:] + '.mov'
        self._path = Path(folder).resolve() / self.filename
        self._path.exists() or self._path.touch()


    def delete(self):
        self._path.unlink()

    @property
    def path(self):
        return str(self._path)



def download_media(bot, url, file_path):
        response = bot.session.get(url, stream=True)
        if response.status_code == 200:
                with open(file_path, 'wb') as f:
                    for chunk in response:
                        f.write(chunk)
        return str(Path(file_path).resolve())


def media_type(media):
    if media['media_type'] == 1:
        return 'photo'
    elif media['media_type'] == 2:
        return 'video'
    else:
        first_media = media["carousel_media"][0]["media_type"]
        return media_type(first_media)



from funcy import rcompose, take
import random
from ..nodes import Node
from .common import accepts


@accepts(Node)
def shuffle(bot, nodes,  args) -> Node:

    amount = args.get('amount')
    if amount:
        nodes = take(amount, nodes)
        if amount <= len(nodes):
            nodes = random.sample(nodes, k=amount)
        else:
            nodes = random.sample(nodes, k=len(nodes))
    else:
        nodes = list(nodes)
        nodes = random.sample(nodes, k=len(nodes))

    return nodes, bot.last



from .node import Node
from .common import attributes




class Arg(Node):

    def __init__(self, *, generic=None, value=None):
        self.value = value

        if generic:
            self.value = generic

    def __repr__(self):
            return 'Arg(value=\'{}\')'.format(self.value)



import datetime
from pathlib import Path
import dataset
import json
import time
from funcy import partial
from ..api import API
from .settings import DELAY, TOTAL, MAX_PER_DAY
from .predicates import not_in_cache

class Bot:

    id = 0

    def __init__(
                 self,
                 username,
                 password,
                 logs_file=None,
                 cache_file=None,
                 cookie_file=None,
                 proxy=None,
                 device=None):

        self.cache_file = make_cache_file(cache_file, username + '_cache.db')
        self.logs_file = make_logs_file( logs_file, username + '_logs.html')
        self.cookie_file = make_cookie_file(cookie_file, username + '_cookie.json')

        self.id = Bot.id
        self.username = username
        Bot.id += 1

        self.predicates = [] # [partial(not_in_cache, self), ]

        self.start_time = datetime.datetime.now()
        self.api = API(logs_file=self.logs_file, id=self.id, username=username, device=device)
        self.logger = self.api.logger

        self.total = TOTAL
        self.delay = DELAY
        self.max_per_day = MAX_PER_DAY

        # methods used in propertis used in yaml
        self._followers_ids = []
        self._followers_usernames = []
        self._following_ids = []
        self._following_usernames = []



        self.api.login(username, password, proxy=proxy,
                       cookie_fname=self.cookie_file)



    def __repr__(self):
        return 'Bot(username=\'{}\', id={})'.format(self.username, self.id)

    @property
    def cache(self):
        return dataset.connect(make_db_url(self.cache_file), engine_kwargs = {'connect_args': {'check_same_thread' : False}})

    @property
    def followers_ids(self):
            if self._followers_ids:
                print(self._followers_ids)

                return self._followers_ids
            else:
                data = cycled_api_call(99999, self, self.api.get_user_followers, id, 'users')
                user_ids = map(lambda item: item['pk'], data)
                self._followers_ids = list(user_ids)
                print(self._followers_ids)
                return self._followers_ids


    @property
    def last(self):
        if self.api.last_json:
            return self.api.last_json
        else:
            return {}

    def reached_limit(self, key):
        current_date = datetime.datetime.now()
        passed_days = (current_date.date() - self.start_time.date()).days
        if passed_days > 0:
            self._reset_counters()
        return self.max_per_day[key] - self.total[key] < 0



    def filter(self, nodes):
        """
        this filters every node before an interaction,
        you can customize this bychanging the predicates.
        prediacte is a function that takes a node as argument
        and returns a boolean
        """
        for predicate in self.predicates:
            nodes = filter(predicate, nodes)

        return nodes

    def suitable(self, node, table=None, specifier=None):
        """
        same as filter but only one node, returns True if node in suitable
        """
        bool = True

        for predicate in self.predicates:
            bool = bool and predicate(
                node,
                table=table,
                specifier=specifier
            )

        return bool


    def sleep(self, type='usual'):

        if type in self.delay:
            self.logger.debug('sleeping for {} seconds'.format(self.delay[type]))
            time.sleep(self.delay[type])
        else:
            self.logger.debug('sleeping for {} seconds'.format(self.delay['usual']))
            time.sleep(self.delay['usual'])


    def _reset_counters(self):
        for k in self.total:
            self.total[k] = 0
        self.start_time = datetime.datetime.now()




def make_db_url( file):
    return 'sqlite:///{}'.format(str(file.resolve()))

def make_logs_file( file, name):
    if not file:
        file = Path(str(Path('.') / '_logs' / name)).resolve()
        file.parent.exists() or file.parent.mkdir()
    file = Path(file)
    file.exists() or file.touch()
    return str(file.resolve())

def make_cache_file( file, name):
    if not file:
        file = Path(str(Path('.') / '_cache' / name)).resolve()
        file.parent.exists() or file.parent.mkdir()
    file = Path(file)
    file.exists() or file.touch()
    return file.resolve()

def make_cookie_file( file, name):
    if not file:
        file = Path(str(Path('.') / '_cookies' / name)).resolve()
        file.parent.exists() or file.parent.mkdir()
    file = Path(file)
    file.exists() or file.touch()
    return str(file.resolve())


def cycled_api_call(amount, bot, api_method, api_argument, key,  ):

    next_max_id = ''
    sleep_track = 0
    done = 0


    while True:
        bot.logger.info('new get cycle with %s' % api_method.__name__)
        try:
            api_method(api_argument, max_id=next_max_id)
            items = bot.last[key] if key in bot.last else []

            if 'next_max_id' not in bot.last:
                yield from items
                done += len(items)
                return

            elif "more_available" in bot.last and not bot.last["more_available"]:
                yield from items
                done += len(items)
                return

            elif "big_list" in bot.last and not bot.last['big_list']:
                yield from items
                done += len(items)
                return

            # elif (done + len(items)) >= max:
            #     yield from items[:(max - done)]
            #     done += len(items)
            #     return

            else:
                yield from items
                done += len(items)

        except Exception as exc:
            bot.logger.error('exception in cycled_api_call: {}'.format(exc))
            yield from []
            return

        if sleep_track > 10:
            bot.logger.debug('sleeping some time while getting')
            bot.sleep('getter')
            sleep_track = 0

        bot.sleep('usual')
        next_max_id = bot.last.get("next_max_id", "")
        sleep_track += 1



# Devices taken from
# https://github.com/mgp25/Instagram-API/blob/master/src/Devices/GoodDevices.php
DEFAULT_DEVICE = 'samsung_galaxy_s7'
DEVICES = {
    # Released on March 2016
    'samsung_galaxy_s7': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 24,
        'android_release': '7.0',
        'dpi': '640dpi',
        'resolution': '1440x2560',
        'manufacturer': 'samsung',
        'device': 'SM-G930F',
        'model': 'herolte',
        'cpu': 'samsungexynos8890',
    },
    # Released on January 2017
    'huawei_mate_9_pro': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 24,
        'android_release': '7.0',
        'dpi': '640dpi',
        'resolution': '1440x2560',
        'manufacturer': 'HUAWEI',
        'device': 'LON-L29',
        'model': 'HWLON',
        'cpu': 'hi3660'
    },
    # Released on November 2016
    'one_plus_3t': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 24,
        'android_release': '7.0',
        'dpi': '380dpi',
        'resolution': '1080x1920',
        'manufacturer': 'OnePlus',
        'device': 'ONEPLUS A3010',
        'model': 'OnePlus3T',
        'cpu': 'qcom'
    },
    # Released on April 2016
    'lg_g5': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 23,
        'android_release': '6.0.1',
        'dpi': '640dpi',
        'resolution': '1440x2392',
        'manufacturer': 'LGE/lge',
        'device': 'RS988',
        'model': 'h1',
        'cpu': 'h1'
    },
    # Released on June 2016
    'zte_axon_7': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 23,
        'android_release': '6.0.1',
        'dpi': '640dpi',
        'resolution': '1440x2560',
        'manufacturer': 'ZTE',
        'device': 'ZTE A2017U',
        'model': 'ailsa_ii',
        'cpu': 'qcom'
    },
    # Released on March 2016
    'samsung_galaxy_s7_edge': {
        'instagram_version': '26.0.0.10.86',
        'android_version': 23,
        'android_release': '6.0.1',
        'dpi': '640dpi',
        'resolution': '1440x2560',
        'manufacturer': 'samsung',
        'device': 'SM-G935',
        'model': 'hero2lte',
        'cpu': 'samsungexynos8890'
    },
}




from typing import List
from funcy import  rcompose, mapcat
from ..nodes import  Media, User
from .common import accepts

@accepts(Media)
def likers(bot, nodes,  args) -> List[Media]:

    pack_user = lambda item: User(id=item['pk'], username=item['username'], data=item)
    amount = args.get('amount')

    process = rcompose(
            lambda media: media.id if media.id else media.get_id(bot),
            lambda id: get_likers(id, bot, amount),
            lambda gen: map(pack_user, gen)
        )

    result = mapcat(process, nodes)


    return result, bot.last





def get_likers(id, bot, amount):
    bot.api.get_media_likers(id)
    if 'users' in bot.last:
        yield from bot.last['users'][:amount]
    else:
        yield from []



from .common import accepts
from ..nodes import Node, User, Media

from .common import today, tap, dotdict
from ..bot import Bot
import json
from dataset import connect
from funcy import rcompose, ignore, mapcat
from itertools import islice
from colorama import init, Fore
import time



@accepts(Node)
def _print(bot: Bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else 1
        model = args['model']

    except KeyError as exc:
        bot.logger.error('please add all necessary args, {}'.format( exc))
        return [], {}


    def process(node):
        """
        model:
            name:      x.full_name
            id:        x.pk
            followers: x.followers_count
        """
        insertion = dotdict()
        for name, expr in model.items():
            insertion[name] = evaluate(expr, node, bot=bot)

        init()
        print()
        print(Fore.CYAN + json.dumps(insertion, indent=4))
        print()
        return node

    max = ignore(OverflowError, None)(lambda: int(max))()
    nodes = map(process, islice(nodes, max))

    return nodes, bot.last



def evaluate(expr, node, bot):
    x = node._data or node.get_data(bot)
    x = dotdict(**x)
    value = xeval(expr, x)
    if not value:
        x = node.get_data(bot)
        x = dotdict(**x)
        return xeval(expr, x)
    else:
        return value

def xeval(expr, x):
    try:
        return eval(expr, dotdict(x=x))

    except (KeyError, AttributeError):
        return None





from .make_task import make_task, partitionate
from .make_bots import make_bots
from .populate import populate_object, populate_string
from .reducer import Reducer
from .threads import start, wait
import traceback

def execute(script, variables={}):

    if type(script) == str:
        script = populate_string(script, variables)
    else:
        script = populate_object(script, variables)

    bots = make_bots(script)

    script_name = script['name'] if script['name'] else 'unnmaed script'

    data = dict()

    try:
        for instruction in script['actions']:
            interaction = list(instruction.keys())[0]

            threads = []
            task = make_task(instruction)

            for (task, bot) in partitionate(task, bots):
                bot.logger.debug('nodes in execute: %s' % task.nodes)
                state = dict(nodes=task.nodes, bot=bot, data=dict(), errors=[])
                # bot.logger.debug(str(bot) + ' ' + str(state))
                edges = [dict(type=edge['type'], args=edge['args']) for edge in task.edges]
                # bot.logger.debug(edges)
                threads += [Reducer(state, edges)]
                bot.logger.debug('new task of type {} and new thread, in script {}'.format(interaction, script_name))
                bot.logger.debug('edges : {}, nodes: {}'.format(edges, list(state['nodes'])))


            threads = start(threads)
            threads = wait(threads)

            data['__' + interaction + '_interaction__'] = [thread.get_data() for thread in threads]
            # {'thread' + thread.name: thread.get_data() for thread in threads}


    except KeyboardInterrupt:
        bot.logger.warn('keyboard interrupt')
        raise

    except Exception as exc:
        print(
            exc.__class__.__name__,
            ':',
            exc,
            '\n',
            '\n'.join(traceback.format_exc().split('\n'))
        )
        raise

    finally:
        return data



from .node import Node
from .common import attributes





class Hashtag(Node):

    def __init__(self, *, generic=None, name=None, data=None):

        self._name = name
        self._data = data

        if generic:
            self._name = generic

    def __repr__(self):
        name, data = attributes(self)

        if name:
            return 'Hashtag(name=\'{}\')'.format(name)
        elif data:
            return 'Hashtag(data=\'{...}\')'

    @property
    def name(self):
        name, data = attributes(self)

        if name:
            return name
        else:
            return False



from typing import List
from funcy import  rcompose, raiser, ignore, mapcat, partial, tap as _tap
import time
from random import choice
from ..bot import Bot
from .common import accepts, today, tap, extract_urls, substitute_vars
from ..nodes import Node, User, Media




@accepts(Media)
def comment(bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else 1
        comments = args['comments']
    except:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}

    count = 0



    def increment():
        bot.total['comments'] += 1
        nonlocal count
        count += 1

    stop = raiser(StopIteration)

    def store_in_cache(node):
        with bot.cache as cache:
            cache['commented'].insert(
                dict(identifier=node.id,
                    specifier=str(comments),
                    time=today(),
                    type='media')
            )
        return node

    return_if_suitable = lambda node: node \
        if bot.suitable(node, table='commented', specifier=str(comments)) \
        else tap(None,lambda: bot.logger.warn('{} not suitable'.format(node)))

    discard_if_reached_limit = lambda node: node \
        if not bot.reached_limit('comments') \
        else tap(None, bot.logger.error('reached commenting daily limit'))

    do_comment_from_groups = lambda node: map(
            lambda cmnts: do_comment(bot, choice(cmnts), node),
            comments) \
         if node else []



    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        return_if_suitable,
        discard_if_reached_limit,
        do_comment_from_groups,
        lambda arr: list(arr)[0] if arr else None,
        lambda node: store_in_cache(node) if node else None,
        lambda x: tap(x, increment) if x else None,
    )


    result = map(process, nodes)
    result = filter(lambda x: x, result)

    return result, bot.last




def do_comment(bot: Bot, text, node, thread_id=None):

    media_id = node.id
    evaluated_text = substitute_vars(text,
        author=ignore(AttributeError, '')(
            lambda: node.get_author(bot).username
        )(),
        caption=node.get_caption(bot),
        geotag=ignore(AttributeError, '')(
            lambda: node.get_geotag(bot).name or ''
        )(),
        usertags=ignore(AttributeError, '')(
            lambda: list(map(lambda x: x.get_username(bot), node.get_usertags(bot))) or []
        )()
    )

    bot.api.comment(
        media_id=media_id,
        comment_text=evaluated_text,
    )
    if bot.last['status'] == 'ok':
        bot.logger.debug('commented %s' % node)
        bot.sleep('comment')
        return node
    else:
        bot.logger.error("comment to {} wasn't posted".format(node))
        bot.sleep('error')
        return None



from typing import List
from funcy import  rcompose, raiser, ignore, mapcat, partial, tap as _tap
import time
from random import choice
from ..bot import Bot
from .common import accepts, today, tap, extract_urls, substitute_vars
from ..nodes import Node, User, Media




@accepts(User)
def text(bot, nodes,  args):


    try:
        max = float(args['max']) if 'max' in args else 1
        messages = args['messages']
    except:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}

    count = 0

    def increment():
        bot.total['texts'] += 1
        nonlocal count
        count += 1

    stop = raiser(StopIteration)

    def store_in_cache(node):
        with bot.cache as cache:
            cache['texted'].insert(
                dict(identifier=node.id,
                    specifier=str(messages),
                    time=today(),
                    type='user')
            )
        return node

    return_if_suitable = lambda node: node \
        if bot.suitable(node, table='texted', specifier=str(messages)) \
        else tap(None,lambda: bot.logger.warn('{} not suitable'.format(node)))

    discard_if_reached_limit = lambda node: node \
        if not bot.reached_limit('texts') \
        else tap(None, bot.logger.error('reached texting daily limit'))

    send_msg_from_groups = lambda node: map(
            lambda msgs: send_message(bot, choice(msgs), node),
            messages) \
         if node else []



    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        # return_if_suitable,
        discard_if_reached_limit,
        send_msg_from_groups,
        lambda arr: list(arr)[0] if arr else None,
        lambda node: store_in_cache(node) if node else None,
        lambda x: tap(x, increment) if x else None,
    )


    result = map(process, nodes)
    result = filter(lambda x: x, result)

    return result, bot.last




def send_message(bot: Bot, text, node, thread_id=None):

    evaluated_text = substitute_vars(text,
        receiver=ignore(AttributeError, '')(
            lambda: node.get_username(bot).username
        )(),
        bio=node.get_bio(bot),
    )

    user_id = node.id if node.id else node.get_id(bot)
    urls = extract_urls(text)
    item_type = 'link' if urls else 'text'

    bot.api.send_direct_item(
        item_type,
        users=[str(user_id)],
        text=evaluated_text,
        thread=thread_id,
        urls=urls
    )
    if bot.last['status'] == 'ok':
        bot.logger.debug('texted %s' % node)
        bot.sleep('text')
        return node
    else:
        bot.logger.error("message to {} wasn't sent".format(node))
        bot.sleep('error')
        return None


# def send_messages(bot, text, user_ids):
#     broken_items = []
#     if not user_ids:
#         bot.logger.info("User must be at least one.")
#         return broken_items
#     bot.logger.info("Going to send %d messages." % (len(user_ids)))
#     for user in tqdm(user_ids):
#         if not bot.send_message(text, user):
#             bot.error_delay()
#             broken_items = user_ids[user_ids.index(user):]
#             break
#     return broken_items



from typing import List
from funcy import rcompose, ignore, tap as _tap, partial, mapcat
import time
from ..bot import Bot
from ..nodes import User, Media
from .common import accepts, cycled_api_call


@accepts(User)
def followers(bot: Bot, nodes,  args) -> List[User]:

    # bot.logger.debug('nodes at followers %s' % list(nodes)[:3])
    #
    # nodes = iter(list(nodes))
    amount = args.get('amount')

    pack_user = lambda item: User(id=item['pk'], username=item['username'], data=item)

    process = rcompose(
            lambda user: user.id if user.id else user.get_id(bot),
            lambda id: cycled_api_call(amount, bot, bot.api.get_user_followers, id, 'users'),
            lambda gen: map(pack_user, gen)

        )

    result = mapcat(process, nodes)

    return result, bot.last



from funcy import wraps, autocurry
from datetime import datetime
import time
from typing import List
from ..nodes import Node
import re


def accepts(Class):

    def _accepts(original):

        @wraps(original)
        def enhanced(bot, nodes,  args):

            # print('nodes in accepts for {}: {}'.format(original.__name__, nodes))

            result = original(bot, nodes,  args)

            return result

        enhanced.accepts = Class

        return enhanced

    return _accepts





def today():
    return datetime.now().strftime("%Y-%m-%d")

def parse_date(date):
    return datetime.strptime(date, "%Y-%m-%d" )





def cycled_api_call(amount, bot, api_method, api_argument, key):

    amount = amount or 9999

    next_max_id = ''
    sleep_track = 0
    done = 0


    while True:
        bot.logger.info('new get cycle with %s' % api_method.__name__)
        try:
            api_method(api_argument, max_id=next_max_id)
            items = bot.last[key] if key in bot.last else []
            size = len(items)

            if any([
                'next_max_id' not in bot.last,
                "more_available" in bot.last and not bot.last["more_available"],
                "big_list" in bot.last and not bot.last['big_list']
            ]):
                yield from items[:amount - done]
                done += size
                return

            # elif (done + size) >= max:
            #     yield from items[:(max - done)]
            #     done += size
            #     return

            else:
                yield from items[:amount - done]
                done += size
                if done > amount:
                    return

        except Exception as exc:
            bot.logger.error('exception in cycled_api_call: {}'.format(exc))
            yield from []
            return

        if sleep_track > 10:
            bot.logger.debug('sleeping some time while getting')
            bot.sleep('getter')
            sleep_track = 0

        bot.sleep('usual')
        next_max_id = bot.last.get("next_max_id", "")
        sleep_track += 1


def tap(x, lazy_fun):
    lazy_fun()
    return x


def propagate(exc):
    raise exc



class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def extract_urls(text):
    url_regex = r'((?:(?:http|https|Http|Https|rtsp|Rtsp)://(?:(?:[a-zA-Z0-9$\-\_\.\+\!\*\'\(\)\,\;\?\&\=]|(?:%[a-fA-F0-9]{2})){1,64}(?::(?:[a-zA-Z0-9$\-\_\.\+\!\*\'\(\)\,\;\?\&\=]|(?:%[a-fA-F0-9]{2})){1,25})?@)?)?(?:(?:(?:[a-zA-Z0-9\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF\_][a-zA-Z0-9\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF\_\-]{0,64}\.)+(?:(?:aero|arpa|asia|a[cdefgilmnoqrstuwxz])|(?:biz|b[abdefghijmnorstvwyz])|(?:cat|com|coop|c[acdfghiklmnoruvxyz])|d[ejkmoz]|(?:edu|e[cegrstu])|f[ijkmor]|(?:gov|g[abdefghilmnpqrstuwy])|h[kmnrtu]|(?:info|int|i[delmnoqrst])|(?:jobs|j[emop])|k[eghimnprwyz]|l[abcikrstuvy]|(?:mil|mobi|museum|m[acdeghklmnopqrstuvwxyz])|(?:name|net|n[acefgilopruz])|(?:org|om)|(?:pro|p[aefghklmnrstwy])|qa|r[eosuw]|s[abcdeghijklmnortuvyz]|(?:tel|travel|t[cdfghjklmnoprtvwz])|u[agksyz]|v[aceginu]|w[fs]|(?:\u03B4\u03BF\u03BA\u03B9\u03BC\u03AE|\u0438\u0441\u043F\u044B\u0442\u0430\u043D\u0438\u0435|\u0440\u0444|\u0441\u0440\u0431|\u05D8\u05E2\u05E1\u05D8|\u0622\u0632\u0645\u0627\u06CC\u0634\u06CC|\u0625\u062E\u062A\u0628\u0627\u0631|\u0627\u0644\u0627\u0631\u062F\u0646|\u0627\u0644\u062C\u0632\u0627\u0626\u0631|\u0627\u0644\u0633\u0639\u0648\u062F\u064A\u0629|\u0627\u0644\u0645\u063A\u0631\u0628|\u0627\u0645\u0627\u0631\u0627\u062A|\u0628\u06BE\u0627\u0631\u062A|\u062A\u0648\u0646\u0633|\u0633\u0648\u0631\u064A\u0629|\u0641\u0644\u0633\u0637\u064A\u0646|\u0642\u0637\u0631|\u0645\u0635\u0631|\u092A\u0930\u0940\u0915\u094D\u0937\u093E|\u092D\u093E\u0930\u0924|\u09AD\u09BE\u09B0\u09A4|\u0A2D\u0A3E\u0A30\u0A24|\u0AAD\u0ABE\u0AB0\u0AA4|\u0B87\u0BA8\u0BCD\u0BA4\u0BBF\u0BAF\u0BBE|\u0B87\u0BB2\u0B99\u0BCD\u0B95\u0BC8|\u0B9A\u0BBF\u0B99\u0BCD\u0B95\u0BAA\u0BCD\u0BAA\u0BC2\u0BB0\u0BCD|\u0BAA\u0BB0\u0BBF\u0B9F\u0BCD\u0B9A\u0BC8|\u0C2D\u0C3E\u0C30\u0C24\u0C4D|\u0DBD\u0D82\u0D9A\u0DCF|\u0E44\u0E17\u0E22|\u30C6\u30B9\u30C8|\u4E2D\u56FD|\u4E2D\u570B|\u53F0\u6E7E|\u53F0\u7063|\u65B0\u52A0\u5761|\u6D4B\u8BD5|\u6E2C\u8A66|\u9999\u6E2F|\uD14C\uC2A4\uD2B8|\uD55C\uAD6D|xn--0zwm56d|xn--11b5bs3a9aj6g|xn--3e0b707e|xn--45brj9c|xn--80akhbyknj4f|xn--90a3ac|xn--9t4b11yi5a|xn--clchc0ea0b2g2a9gcd|xn--deba0ad|xn--fiqs8s|xn--fiqz9s|xn--fpcrj9c3d|xn--fzc2c9e2c|xn--g6w251d|xn--gecrj9c|xn--h2brj9c|xn--hgbk6aj7f53bba|xn--hlcj6aya9esc7a|xn--j6w193g|xn--jxalpdlp|xn--kgbechtv|xn--kprw13d|xn--kpry57d|xn--lgbbat1ad8j|xn--mgbaam7a8h|xn--mgbayh7gpa|xn--mgbbh1a71e|xn--mgbc0a9azcg|xn--mgberp4a5d4ar|xn--o3cw4h|xn--ogbpf8fl|xn--p1ai|xn--pgbs0dh|xn--s9brj9c|xn--wgbh1c|xn--wgbl6a|xn--xkc2al3hye2a|xn--xkc2dl3a5ee0h|xn--yfro4i67o|xn--ygbi2ammx|xn--zckzah|xxx)|y[et]|z[amw]))|(?:(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[1-9][0-9]|[1-9])\.(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[1-9][0-9]|[1-9]|0)\.(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[1-9][0-9]|[1-9]|0)\.(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[1-9][0-9]|[0-9])))(?::\d{1,5})?(?:/(?:(?:[a-zA-Z0-9\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF\;\/\?\:\@\&\=\#\~\-\.\+\!\*\'\(\)\,\_])|(?:%[a-fA-F0-9]{2}))*)?)(?:\b|$)'
    urls = re.findall(url_regex, text)

    return urls


effify = lambda non_f_str: f'{non_f_str}'
substitute_vars = lambda txt, **kwds: effify(txt).format(**kwds)



from pathlib import Path
import json
import subprocess
from random import random



def unmask(obj):
    data = json.dumps(obj)
    temp_file = temporary_file('temp_' + str(random())[:2], data)

    with temp_file.open('w') as file:
        file.write(data)

    cmd = 'cat {0!s} | unmask-json --stream --raw'.format(temp_file)

    ps = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    out = ps.communicate()[0]

    temp_file.unlink()

    # out = subprocess.check_output(
    #     [cmd ], shell=True, stderr=subprocess.PIPE)

    return out.decode('ascii')



def temporary_file(name, content):

    folder = Path(__file__).parent

    file = Path(folder.resolve() / (name + '_temporary'))

    file.exists() or file.touch()

    return file.resolve()



from .common import accepts
from ..nodes import Node, User, Media

from .common import today, tap
from ..bot import Bot
from funcy import rcompose, raiser, tap as _tap
import time



@accepts(User)
def follow(bot: Bot, nodes,  args):

    max = float(args['max']) if 'max' in args else 1
    count = 0

    def increment():
        bot.total['follows'] += 1
        nonlocal count
        count += 1

    stop = raiser(StopIteration)

    process = rcompose(
        # lambda x: tap(x, lambda: bot.logger.warn('{}._data: \n{}'.format(x, unmask(x._data)))),
        lambda x: stop() if x and count >= max else x,
        # lambda node: node \
        #     if bot.suitable(node) \
        #     else tap(None,lambda: bot.logger.warn('{} not suitable'.format(node))),
        lambda node: follow_user(node, bot=bot) \
            if node else None,
        lambda x: tap(x, increment) if x else None,
    )


    followed = map(process, nodes)
    followed = filter(lambda x: x, followed)

    return followed, bot.last



def follow_user(user, bot):
    bot.api.follow(user.id)
    if bot.last['status'] != 'ok':
        bot.logger.warn('request didn\'t return "ok" following {}'.format(user.username))
        bot.sleep('error')
        return None
    else:
        with bot.cache as cache:
            cache['followed'].insert(dict(identifier=user.id, time=today(), type='user',))

        bot.logger.info('followed %s' % user)
        bot.sleep('follow')
        return user
